bl_info = {  
    "name": "REDHALO_IMPORT_MAX_SCENE",
    "author": "Red Halo Studio",
    "version": (0, 9),
    "blender": (2, 80, 0),
    "location": "View 3D > Tools > Red Halo Tools",
    "description": "导入3ds Max场景",
    "wiki_url": "",
    "tracker_url": "",
    "category": "Import"
 }

from math import pi, radians
import bpy, bpy_extras, colorsys, tempfile, os, traceback
from bpy.types import Operator, Operator, AddonPreferences
from bpy.props import StringProperty, BoolProperty, EnumProperty
from mathutils import Matrix, Vector
from xml.etree import ElementTree as xml
from winreg import *
from bpy_extras.io_utils import ImportHelper
from struct import *

__version__ = "0.9.1"

def FixDaeName(string):
    # fbxname = "FBXASC231FBXASC170FBXASC151FBXASC229FBXASC184FBXASC152"
    names = [chr(int(i)) for i in string.split("FBXASC")[1:]]
    str = "".join(names)
    return str.encode("iso-8859-1").decode("utf-8")

## 检测Max文件版本号 
def FindMax(filepath):
    version = ""
    sv = ""
    max_version = {2009: 10, 2010: 11, 2010:12, 2011:13, 2012:14, 2013:15, 2014:16, 2015:17, 2016:18, 2017:19, 2018:20, 2019:21, 2020:22, 2021:23, 2022:24, 2023:25, 2024:26, 2025:27, 2026:28, 2027:29, 2028:30}
    vr_v = 10000
    with open(filepath, "rb") as f:
        buff = f.read()
        # Save Max Version
        save_bit = (b"\x33\x64\x73\x20\x4D\x61\x78\x20\x56\x65\x72\x73\x69\x6F\x6E\x3A", b"\x33\x00\x64\x00\x73\x00\x20\x00\x4D\x00\x61\x00\x78\x00\x20\x00\x56\x00\x65\x00\x72\x00\x73\x00\x69\x00\x6F\x00\x6E\x00\x3A\x00", b"\x33\x64\x73\x20\x4D\x61\x78\x20\xE7\x89\x88\xE6\x9C\xAC\x3A\x20")
        save_bit_offset = (36, 48, 24)
        for s in zip(save_bit, save_bit_offset):
            fi = buff.find(s[0])
            if fi != -1:
                f.seek(fi)
                version = f.read(s[1]).decode("utf-8", "ignore").replace("\x00","").replace(",", ".").replace(" ","").replace("3dsMaxVersion:", "").replace("3dsMax版本:", "")

        # Save As Version
        SaveAs_bit = (b"\x53\x61\x76\x65\x64\x20\x41\x73", b"\x53\x00\x61\x00\x76\x00\x65\x00\x64\x00\x20\x00\x41\x00\x73\x00\x20", b"\xE5\x8F\xA6\xE5\xAD\x98\xE4\xB8\xBA\xE7\x89\x88\xE6\x9C\xAC\x3A")
        
        # Vrend sting : vrender版本号.dlr
        vrend = (b"x\76x\00x\72x\00x\65x\00x\6Ex\00x\64x\00x\65x\00x\72x\00")

        SaveAs_bit_offset = (28, 48, 24)        
        for s in zip(SaveAs_bit, SaveAs_bit_offset):
            fi = buff.find(s[0])
            if fi != -1:
                f.seek(fi)
                sv = f.read(s[1]).decode("utf-8","ignore").replace("\x00","").replace(",", ".").replace(" ", "").replace("SavedAsVersion:", "").replace("另存为版本:", "")
            else:
                sv = version
        # Vrend sting : vrender版本号.dlr
        vrend = (b"\x76\x00\x72\x00\x65\x00\x6E\x00\x64\x00\x65\x00\x72\x00")
        findIndex = buff.find(vrend)
        if  findIndex != -1: 
            f.seek(findIndex)
            v_version = f.read(32).decode("utf-16")
            vrender_version = int(os.path.splitext(v_version)[0].replace("vrender", ""))
            vr_v = max_version[vrender_version]

    mv = 10000 if version == "" else int(version.split(".")[0])
    av = 10000 if sv == "" else int(sv.split(".")[0])
    vv = min(mv, av, vr_v)
    return vv

def GetImportPath():
    # Reg Name in Windows
    RegRoot = ConnectRegistry(None, HKEY_CURRENT_USER)
    RegSubDir = "SOFTWARE\\REDHALO"

    try:
        RegKey = OpenKey(RegRoot, RegSubDir)
        keyName = EnumValue(RegKey, 0)
        return keyName[1]  
    except:
        return ""          

def Create_ColorCorrection():
    # Color Correction Node
    CC_group = bpy.data.node_groups.new('ColorCorrection_RH', 'ShaderNodeTree')
    group_inputs = CC_group.nodes.new('NodeGroupInput')
    group_inputs.location = (0, 0)
    items =[]
    if bpy.app.version < (4, 00, 0):        
        CC_group.inputs.new('NodeSocketColor','Color')
        CC_group.inputs.new('NodeSocketFloat','Bright')
        CC_group.inputs.new('NodeSocketFloat','Contrast')
        CC_group.inputs.new('NodeSocketFloat','Hue')
        CC_group.inputs.new('NodeSocketFloat','Saturation')
        CC_group.inputs.new('NodeSocketFloat','Invert')
        CC_group.inputs.new('NodeSocketFloat','Gamma')
        items = CC_group.inputs
    else:
        CC_group.interface.new_socket(name="Color", in_out = "INPUT", socket_type="NodeSocketColor")
        CC_group.interface.new_socket(name="Bright", in_out = "INPUT", socket_type="NodeSocketFloat")
        CC_group.interface.new_socket(name="Contrast", in_out = "INPUT", socket_type="NodeSocketFloat")
        CC_group.interface.new_socket(name="Hue", in_out = "INPUT", socket_type="NodeSocketFloat")
        CC_group.interface.new_socket(name="Saturation", in_out = "INPUT", socket_type="NodeSocketFloat")
        CC_group.interface.new_socket(name="Invert", in_out = "INPUT", socket_type="NodeSocketFloat")
        CC_group.interface.new_socket(name="Gamma", in_out = "INPUT", socket_type="NodeSocketFloat")
        items = CC_group.interface.items_tree

    items["Color"].default_value = (1,1,1,1)

    items["Bright"].min_value = -100
    items["Bright"].max_value = 100

    items["Contrast"].min_value = -100
    items["Contrast"].max_value = 100

    items["Hue"].default_value = 0.5
    items["Hue"].min_value = 0
    items["Hue"].max_value = 1

    items["Saturation"].default_value = 1
    items["Saturation"].min_value = 0
    items["Saturation"].max_value = 1

    items["Invert"].default_value = 0
    items["Invert"].min_value = 0
    items["Invert"].max_value = 1

    items["Gamma"].default_value = 1
    items["Gamma"].min_value = 0.01
    items["Gamma"].max_value = 10

    group_outputs = CC_group.nodes.new("NodeGroupOutput")
    group_outputs.location = (1000, 0)
    if bpy.app.version < (4, 00, 0):        
        CC_group.outputs.new("NodeSocketColor", "Color")
    else:
        CC_group.interface.new_socket(name="Color", in_out = "OUTPUT", socket_type="NodeSocketColor")

    BCNode = CC_group.nodes.new("ShaderNodeBrightContrast")
    BCNode.location = (200, 0)
    HSVNode = CC_group.nodes.new("ShaderNodeHueSaturation")
    HSVNode.location = (400, 0)
    InvertNode =  CC_group.nodes.new("ShaderNodeInvert")
    InvertNode.location = (600, 0)
    GammaNode = CC_group.nodes.new("ShaderNodeGamma")
    GammaNode.location = (800, 0)

    CC_group.links.new(group_inputs.outputs["Color"], BCNode.inputs["Color"])
    CC_group.links.new(group_inputs.outputs["Bright"], BCNode.inputs["Bright"])
    CC_group.links.new(group_inputs.outputs["Contrast"], BCNode.inputs["Contrast"])
    CC_group.links.new(group_inputs.outputs["Hue"], HSVNode.inputs["Hue"])
    CC_group.links.new(group_inputs.outputs["Saturation"], HSVNode.inputs["Saturation"])
    CC_group.links.new(group_inputs.outputs["Invert"], InvertNode.inputs["Fac"])
    CC_group.links.new(group_inputs.outputs["Gamma"], GammaNode.inputs["Gamma"])

    CC_group.links.new(BCNode.outputs["Color"], HSVNode.inputs["Color"])
    CC_group.links.new(HSVNode.outputs["Color"], InvertNode.inputs["Color"])
    CC_group.links.new(InvertNode.outputs["Color"], GammaNode.inputs["Color"])
    CC_group.links.new(GammaNode.outputs["Color"], group_outputs.inputs["Color"])

def ConvertMatrix(matrixstr):
    '''
    #### 转换max下的3x4Matrix
    '''
    # Max 输出为 3*4 矩阵
    # (matrix3 [0.766044,0.642788,0] [-0.642788,0.766044,0] [0,0,1] [2306.89,2755.7,0])

    # Blender 输出为 4*4 矩阵
    #<Matrix 4x4 ( 0.7660, -0.6428, 0.0000, 2306.8943)
    #            ( 0.6428,  0.7660, 0.0000, 2755.6978)
    #            (-0.0000, -0.0000, 1.0000,    0.0000)
    #            ( 0.0000,  0.0000, 0.0000,    1.0000)>

    a = matrixstr[9:-1].split()
    matrix = Matrix()
    for i in range(len(a)):
        b = eval(a[i])
        b.append(0) if i < 3 else b.append(1)
        matrix[i] = b

    return matrix.transposed()

def ConvertColor(color):
    _c = (color[7:-1]+" 255").split()
    _clr = [float(c)/255.0 for c in _c]
    return _clr

def CollectImagesNode(node):
    allnodes = []
    allnodes.append(node)    
    inputs = [n for n in node.inputs if n.is_linked]
    
    for input in inputs:
        for l in input.links:
            allnodes.extend(CollectImagesNode(l.from_node))
            
    return allnodes

def CreateNode(nodes, links, xmlPath, ParentNode, texmap_type):
    '''
    #### nodes      :节点
    #### links      :连接
    #### xmlPath    :查找路径
    #### ParentNode :父级节点
    #### type       :节点类型
    '''
    texmap = xmlPath    
    _node = texmap.findall("./")
    
    # Bitmap 已完成
    if texmap_type in ("Bitmap", "VRayBitmap") :
        bitmap_path = _node[0].text
        if bitmap_path != "":            
            # Scale
            uTile = float(texmap.attrib["U_Tiling"])
            vTile = float(texmap.attrib["V_Tiling"])
            # Location
            uOffset = float(texmap.attrib["U_Offset"])
            vOffset = float(texmap.attrib["V_Offset"])
            #Rotation
            wAngle = float(texmap.attrib["W_Angle"])
            vAngle = float(texmap.attrib["V_Angle"])
            uAngle = float(texmap.attrib["U_Angle"])
            # Mapping
            mapping = int(texmap.attrib["Mapping"])
            mappingtype = int(texmap.attrib["MappingType"])
            
            ImageShader = CreateImageNode(nodes, links, bitmap_path, position=(uOffset, vOffset, 0), rotation=(radians(uAngle), radians(vAngle), radians(wAngle)), scale=(uTile, vTile, 1), mapping=mapping, mappingtype=mappingtype)     

            if ImageShader:
                links.new(ImageShader.outputs["Color"], ParentNode)
    
    # 平铺 部分完成
    elif texmap_type == "Tiles" :
        BrickShader = nodes.new("ShaderNodeTexBrick")
        # BrickShader.location = p_loc - Vector((p_width + 100, 0))
        links.new(BrickShader.outputs["Color"], ParentNode)

        # 平铺纹理基础参数设置
        Brick_Color = texmap.attrib["Brick_color"]
        Mortar_Color = texmap.attrib["Mortar_color"]
        BrickShader.inputs["Color1"].default_value = ConvertColor(Brick_Color)
        BrickShader.inputs["Color2"].default_value = ConvertColor(Brick_Color)
        BrickShader.inputs["Mortar"].default_value = ConvertColor(Mortar_Color)

        BrickShader.inputs["Scale"].default_value = 1
        BrickShader.inputs["Mortar Size"].default_value = float(texmap.attrib["Horizontal_Gap"]) / 100
        BrickShader.inputs["Brick Width"].default_value = 1 / float(texmap.attrib["Horizontal_Count"])
        BrickShader.inputs["Row Height"].default_value = 1 / float(texmap.attrib["Vertical_Count"])

        Brick_Type = texmap.attrib["Tile_Type"]
        Brick_offset = 0.5
        if Brick_Type == "5" :
            Brick_offset = 0
        elif Brick_Type == "0" :
            Brick_offset = texmap.attrib["Line_Shift"]
        
        BrickShader.offset = Brick_offset
        
        CreateUVMapping(nodes, links, BrickShader)

        # 判断分缝线有没有纹理，如果有创建相关节点
        if texmap.attrib["Mortar_Map"] != "undefined" :
            MortarNode = _node[0][0]
            MortarType = texmap.attrib["Mortar_Map"].split(":")[-1]
            CreateNode(nodes, links, MortarNode, BrickShader.inputs["Mortar"], MortarType)
        
        # 判断平铺有没有纹理，如果有创建相关节点
        if texmap.attrib["Bricks_Map"] != "undefined" :
            TileNode = _node[1][0]
            TileType = texmap.attrib["Bricks_Map"].split(":")[-1]
            CreateNode(nodes, links, TileNode, BrickShader.inputs["Color1"], TileType)
            CreateNode(nodes, links, TileNode, BrickShader.inputs["Color2"], TileType)

    # 混合 已完成
    elif texmap_type in ("Mix", "RGB_Multiply", "CoronaMix", "RGB_Multiply", "VRayCompTex") :
        #  #### Params : Color1 Color2 MixAmount Texmap1 Texmap2 TexmapMask MixMode
        #  MixMode:
        #  ADD SUBTRACT MULTIPLY DIVIDE MINIMUM MAXIMUM MIX GAMMA DIFFERENCE SCREEN OVERLAY COLORDODGE COLORBURN LINEARBURN LINEARLIGHT SOFTLIGHT HARDLIGHT VIVIDLIGHT PINLIGHT HARDMIX EXCLUSION
        
        # 旧版MixRGB
        MixShader = nodes.new("ShaderNodeMixRGB")
        
        links.new(MixShader.outputs["Color"], ParentNode)

        # 基础参数设置
        amount = texmap.attrib["MixAmount"]
        color1 = texmap.attrib["Color1"]
        color2 = texmap.attrib["Color2"]
        MixShader.inputs["Fac"].default_value = float(amount)/100
        MixShader.inputs["Color1"].default_value = ConvertColor(color1)
        MixShader.inputs["Color2"].default_value = ConvertColor(color2)
        MixShader.blend_type = texmap.attrib["MixMode"]

        if texmap.attrib["Texmap1"] != "undefined":
            map1Node = _node[0][0]
            map1Type = texmap.attrib["Texmap1"].split(":")[-1]
            CreateNode(nodes, links, map1Node, MixShader.inputs["Color1"], map1Type)

        if texmap.attrib["Texmap2"] != "undefined":
            map2Node = _node[1][0]
            map2Type = texmap.attrib["Texmap2"].split(":")[-1]
            CreateNode(nodes, links, map2Node, MixShader.inputs["Color2"], map2Type)

        if texmap.attrib["TexmapMask"] != "undefined":
            maskNode = _node[2][0]
            MaskType = texmap.attrib["TexmapMask"].split(":")[-1]
            CreateNode(nodes, links, maskNode, MixShader.inputs["Fac"], MaskType)

    # 棋盘格  已完成
    elif texmap_type == "Checker":
        CheckerShader = nodes.new("ShaderNodeTexChecker")
        links.new(CheckerShader.outputs["Color"], ParentNode)

        #基础参数
        color1 = texmap.attrib["color1"]
        color2 = texmap.attrib["color2"]
        map1 = texmap.attrib["map1"]
        map2 = texmap.attrib["map2"]

        CheckerShader.inputs["Color1"].default_value = ConvertColor(color1)
        CheckerShader.inputs["Color2"].default_value = ConvertColor(color2)
        CheckerShader.inputs["Scale"].default_value = 2

        if map1 != "undefined" :
            map1Node = _node[0][0]
            map1Type = map1.split(":")[-1]
            CreateNode(nodes, links, map1Node, CheckerShader.inputs["Color1"], map1Type)

        if map2 != "undefined" :
            map2Node = _node[1][0]
            map2Type = map2.split(":")[-1]
            CreateNode(nodes, links, map2Node, CheckerShader.inputs["Color2"], map2Type)
        
        CreateUVMapping(nodes, links, CheckerShader)

    # 衰减  完成IOR部分，其它类型还没找到相应实现方式
    elif texmap_type == "Falloff":        
        # 旧版MixRGB
        MixRGBShader = nodes.new("ShaderNodeMixRGB")
        links.new(MixRGBShader.outputs["Color"], ParentNode)

        color1 = texmap.attrib["color1"]
        color2 = texmap.attrib["color2"]
        MixRGBShader.inputs["Color1"].default_value = ConvertColor(color1)
        MixRGBShader.inputs["Color2"].default_value = ConvertColor(color2)

        # Map1
        if texmap.attrib["map1"] != "undefined":
            map1Node = _node[0][0]
            map1Type = texmap.attrib["map1"].split(":")[-1]
            CreateNode(nodes, links, map1Node, MixRGBShader.inputs["Color1"], map1Type)
        # Map2
        if texmap.attrib["map2"] != "undefined":
            map2Node = _node[1][0]
            map2Type = texmap.attrib["map2"].split(":")[-1]
            CreateNode(nodes, links, map2Node, MixRGBShader.inputs["Color2"], map2Type)

        Falloff_Type = texmap.attrib["type"]
        if Falloff_Type == "2":
            FresnelShader = nodes.new("ShaderNodeFresnel")
            links.new(FresnelShader.outputs["Fac"], MixRGBShader.inputs["Fac"])
            FresnelShader.inputs["IOR"].default_value = float(texmap.attrib["ior"])
        else:
            LayerWeightNode = nodes.new("ShaderNodeLayerWeight")
            links.new(LayerWeightNode.outputs["Facing"], MixRGBShader.inputs["Fac"])

    # 颜色校正节点 已完成
    elif texmap_type in ("Color Correction", "CoronaColorCorrect"):
        # ### Params:
        # #### Color Map Hue Saturation Brightness Contrast Gamma ColorMode
        ###### ColorMode --> NORAML INVERT MONO
        cc_group = nodes.new("ShaderNodeGroup")
        # cc_group.location = p_loc - Vector((p_width + 100, 0))
        cc_group.node_tree = bpy.data.node_groups['ColorCorrection_RH']
        links.new(cc_group.outputs["Color"], ParentNode)

        cc_group.inputs["Color"].default_value = ConvertColor(texmap.attrib["Color"])
        
        if texmap.attrib["ColorMode"] == "MONO":
            cc_group.inputs["Saturation"].default_value = 0
        elif texmap.attrib["ColorMode"] == "INVERT":
            cc_group.inputs["Invert"].default_value = 1

        cc_group.inputs["Bright"].default_value = float(texmap.attrib["Brightness"])/100.0
        cc_group.inputs["Contrast"].default_value = float(texmap.attrib["Contrast"])/100.0
        cc_group.inputs["Gamma"].default_value = 1 / float(texmap.attrib["Gamma"])

        cc_group.inputs["Hue"].default_value = float(texmap.attrib["Hue"])/360.0 + 0.5
        cc_group.inputs["Saturation"].default_value = float(texmap.attrib["Saturation"])/100.0 + 1

        if texmap.attrib["Map"] != "undefined" :
            mapNode = _node[0][0]
            mapType = texmap.attrib["Map"].split(":")[-1]
            CreateNode(nodes, links, mapNode, cc_group.inputs["Color"], mapType)

    # 渐变节点,Blender不支持贴图，所以只保留颜色
    elif texmap_type == "Gradient" :
        GradientNode = nodes.new("ShaderNodeTexGradient")
        ColorRampNode = nodes.new("ShaderNodeValToRGB")

        links.new(GradientNode.outputs["Fac"], ColorRampNode.inputs["Fac"])
        links.new(ColorRampNode.outputs["Color"], ParentNode)
        
        if texmap.attrib["gradientType"] == "1":
            GradientNode.gradient_type = "SPHERICAL"
            CreateUVMapping(nodes, links, GradientNode, position=(-0.5, -0.5, -0.86))
        else:
            CreateUVMapping(nodes, links, GradientNode, position=(-0.5, -0.5, 0))

        ColorRampNode.color_ramp.elements.new(float(texmap.attrib["color2Pos"]))

        ColorRampNode.color_ramp.elements[0].color = ConvertColor(texmap.attrib["color1"])
        ColorRampNode.color_ramp.elements[1].color = ConvertColor(texmap.attrib["color2"])
        ColorRampNode.color_ramp.elements[2].color = ConvertColor(texmap.attrib["color3"])        

    # 渐变坡度节点
    elif texmap_type == "Gradient Ramp" :
        pass
        GradientNode = nodes.new("ShaderNodeTexGradient")
        ColorRampNode = nodes.new("ShaderNodeValToRGB")

        links.new(GradientNode.outputs["Fac"], ColorRampNode.inputs["Fac"])
        links.new(ColorRampNode.outputs["Color"], ParentNode)

    # AO
    elif texmap_type in ("VRayDirt", "CoronaAO") : 
        # Params Radius OccludedColor UnoccludedColor Subdivs OnlySameObject TexmapRadius TexmapOccluded TexmapUnoccluded Mode
        # Mode : OUTSIDE INSIDE BOTH
        MixShader = nodes.new("ShaderNodeMixRGB")

        AONode = nodes.new("ShaderNodeAmbientOcclusion")
        AONode.location = MixShader.location - Vector((MixShader.width + 100, -100))

        links.new(AONode.outputs["AO"], MixShader.inputs["Fac"])
        links.new(MixShader.outputs["Color"], ParentNode)

        AONode.inputs["Color"].default_value = ConvertColor(texmap.attrib["OccludedColor"])
        AONode.samples = int(texmap.attrib["Subdivs"])
        AONode.inputs["Distance"].default_value = float(texmap.attrib["Radius"])

        MixShader.inputs["Color1"].default_value = (0, 0, 0, 1)
        MixShader.inputs["Color2"].default_value = (1, 1, 1, 1)

        if texmap.attrib["OnlySameObject"] == "false" :
            AONode.only_local = False
        else:
            AONode.only_local = True
        
        if texmap.attrib["Mode"] == "INSIDE":
            AONode.inside = True
        else:
            AONode.inside = False

        if texmap.attrib["TexmapOccluded"] != "undefined":
            OccludedMap = _node[0][0]
            OccludedType = texmap.attrib["TexmapOccluded"].split(":")[-1]
            CreateNode(nodes, links, OccludedMap, MixShader.inputs["Color1"], OccludedType)
        
        if texmap.attrib["TexmapUnoccluded"] != "undefined":
            UnoccludedMap = _node[1][0]
            UnoccludedType = texmap.attrib["TexmapUnoccluded"].split(":")[-1]
            CreateNode(nodes, links, UnoccludedMap, MixShader.inputs["Color2"], UnoccludedType)

        if texmap.attrib["TexmapRadius"] != "undefined" :
            RadiusMap = _node[2][0]
            RadiusType = texmap.attrib["TexmapRadius"].split(":")[-1]
            CreateNode(nodes, links, RadiusMap, AONode.inputs[1], RadiusType)

    # 法线/凹凸
    elif texmap_type in ("VRayColor2Bump", "VRayNormalMap", "CoronaNormal", "VRayBump2Normal", "CoronaBumpConverter") :
        '''
        #### Params : NormalMap BumpMap BumpStrength NormalStrength FlipRed FilpGreen SweepRedGreen
        '''
        HasBump = True if texmap.attrib["BumpMap"] != "undefined" else False
        HasNormal = True if texmap.attrib["NormalMap"] != "undefined" else False

        if HasNormal :
            NormalNode = nodes.new("ShaderNodeNormalMap")
            NormalNode.inputs["Strength"].default_value = 1.0 # float(texmap.attrib["NormalStrength"])

            normal_map = texmap[0][0]
            normal_map_type = texmap.attrib["NormalMap"].split(":")[-1]
            CreateNode(nodes, links, normal_map, NormalNode.inputs["Color"], normal_map_type)

        if HasBump :
            BumpNode = nodes.new("ShaderNodeBump")
            BumpNode.inputs["Strength"].default_value = 0.1 #float(texmap.attrib["BumpStrength"])

            Bump_Map = texmap[1][0]
            Bump_Map_Type = texmap.attrib["BumpMap"].split(":")[-1]
            CreateNode(nodes, links, Bump_Map, BumpNode.inputs["Height"], Bump_Map_Type)
        

        ### 如果有凹凸并且有法线--全
        ### 如果有凹凸无法线 --- 只连bump
        ### 如果没有凹凸只有法线  --- 直连Normal
        if HasBump:
            if HasNormal: ## +B + N
                links.new(NormalNode.outputs["Normal"], BumpNode.inputs["Normal"])
                links.new(BumpNode.outputs["Normal"], ParentNode)
            else:  ## +B - N
                links.new(BumpNode.outputs["Normal"], ParentNode)
        else:
            ## -B +N
            if HasNormal:
                links.new(NormalNode.outputs["Normal"], ParentNode)
            #else: ## -B -N
            #    links.new(BumpNode.outputs["Normal"], ParentNode)

    elif texmap_type == "Vertex Color" :
        VCNode = nodes.new("ShaderNodeAttribute")
        if bpy.app.version < (3, 4, 0):
            VCNode.attribute_name = "Col"
        else:
            VCNode.attribute_name = "Attribute"

        links.new(VCNode.outputs["Color"], ParentNode)
    
    elif texmap_type in ("CoronaFrontBack"):
        
        # 旧版MixRGB
        MixShader = nodes.new("ShaderNodeMixRGB")
        links.new(MixShader.outputs["Color"], ParentNode)

        # 基础参数设置
        color1 = texmap.attrib["frontColor"]
        color2 = texmap.attrib["backColor"]
        MixShader.inputs["Color1"].default_value = ConvertColor(color1)
        MixShader.inputs["Color2"].default_value = ConvertColor(color2)

        ## Object Backing Node
        Backfacing_Node = nodes.new("ShaderNodeNewGeometry")
        links.new(Backfacing_Node.outputs["Backfacing"], MixShader.inputs["Fac"])

        if texmap.attrib["frontTexmap"] != "undefined":
            map1Node = _node[0][0]
            map1Type = texmap.attrib["frontTexmap"].split(":")[-1]
            CreateNode(nodes, links, map1Node, MixShader.inputs["Color1"], map1Type)

        if texmap.attrib["backTexmap"] != "undefined":
            map2Node = _node[1][0]
            map2Type = texmap.attrib["backTexmap"].split(":")[-1]
            CreateNode(nodes, links, map2Node, MixShader.inputs["Color2"], map2Type)

    elif texmap_type in ("Composite"):
        #Blender Mode
        blendMode = ('MIX', 'DARKEN', 'MULTIPLY', 'BURN', 'LIGHTEN', 'SCREEN', 'DODGE', 'ADD', 'OVERLAY', 'SOFT_LIGHT', 'LINEAR_LIGHT', 'DIFFERENCE', 'EXCLUSION', 'SUBTRACT', 'DIVIDE', 'HUE', 'SATURATION', 'COLOR', 'VALUE')        
        
        OutNode = None
        for i,layer in enumerate(texmap):
            mode = layer.attrib["Mode"]

            opacity = float(layer.attrib["Opacity"])
            layer_map = layer.attrib["Layer"]
            layer_mask = layer.attrib["Mask"]

            if layer_map != "undefined":
                map1Node = layer[0][0]
                map1Type = layer_map.split(":")[-1]

                if i == 1:
                    # 旧版MixRGB
                    MixShader = nodes.new("ShaderNodeMixRGB")
                    # First
                    layer_1= texmap[0]
                    layer_1_map = layer_1[0][0]
                    layer_1_type = layer_1.attrib["Layer"].split(":")[-1]
                    CreateNode(nodes, links, layer_1_map, MixShader.inputs["Color1"], layer_1_type)
                    # Sceond
                    CreateNode(nodes, links, map1Node, MixShader.inputs["Color2"], map1Type)
                    MixShader.blend_type = mode if mode in blendMode else 'MIX'
                    MixShader.inputs["Fac"].default_value = opacity

                    if layer_mask != "undefined":
                        maskNode = layer[1][0]
                        maskType = layer_mask.split(":")[-1]
                        CreateNode(nodes, links, maskNode, MixShader.inputs["Fac"], maskType)

                    OutNode = MixShader
                elif i > 1: 
                    # 旧版MixRGB
                    MixShader = nodes.new("ShaderNodeMixRGB")
                    links.new(OutNode.outputs["Color"], MixShader.inputs["Color1"])
                    CreateNode(nodes, links, map1Node, MixShader.inputs["Color2"], map1Type)
                    MixShader.blend_type = mode if mode in blendMode else 'MIX'
                    MixShader.inputs["Fac"].default_value = opacity

                    if layer_mask != "undefined":
                        maskNode = layer[1][0]
                        maskType = layer_mask.split(":")[-1]
                        CreateNode(nodes, links, maskNode, MixShader.inputs["Fac"], maskType)

                    OutNode = MixShader
        
        links.new(OutNode.outputs["Color"], ParentNode)

def CreateUVMapping(nodes, links, ParentNode, Coords="UV", position=(0, 0, 0), rotation=(0, 0, 0), scale=(1, 1, 1)):
    '''
    #### Coords      :坐标,默认为 “UV”，参数列表:Generated, Normal, UV, Object, Camera, Window, Reflection
    #### position    :移动，默认为(0,0,0)
    #### rotation    :旋转角度
    #### scale       :缩放
    #### ParentNode  :父级节点
    '''
    mapShader = nodes.new("ShaderNodeMapping")
    links.new(mapShader.outputs["Vector"], ParentNode.inputs["Vector"])
    
    mapShader.location = ParentNode.location - Vector((ParentNode.width, 0))

    mapShader.inputs["Location"].default_value = position
    mapShader.inputs["Rotation"].default_value = rotation
    mapShader.inputs["Scale"].default_value = scale

    texcoordShader = nodes.new("ShaderNodeTexCoord")

    links.new(texcoordShader.outputs[Coords], mapShader.inputs["Vector"])

def CreateImageNode(nodes, links, bitmap, Coors="UV", position=(0, 0, 0), rotation=(0, 0, 0), scale=(1, 1, 1), mapping=0, mappingtype=0):
    '''
    #### bitmap      :图像路径
    #### Coords      :坐标,默认为 “UV”，参数列表:Generated, Normal, UV, Object, Camera, Window, Reflection
    #### position    :移动，默认为(0,0,0)
    #### rotation    :旋转角度
    #### scale       :缩放
    #### ParentNode  :父级节点
    '''

    _, filename = os.path.split(bitmap)
    img = bpy_extras.image_utils.load_image(bitmap, check_existing=True)    
    ImageShader = None
    if img:        
        if mappingtype == 1:
            # Sphere Environ
            if mapping == 0:
                ImageShader = nodes.new("ShaderNodeTexEnvironment")
        else:
            ImageShader = nodes.new("ShaderNodeTexImage")
    
        img.name = filename
        ImageShader.image = img

        if mappingtype == 1:
            # Sphere Environ
            if mapping == 0:
                CreateUVMapping(nodes, links, ImageShader, "Generated", position, rotation, scale)
            if mapping == 3:
                CreateUVMapping(nodes, links, ImageShader, "Window", position, rotation, scale)
        else:
            if position != (0, 0, 0) or rotation != (0, 0, 0) or scale != (1, 1, 1):
                CreateUVMapping(nodes, links, ImageShader, "UV", position, rotation, scale)

    return ImageShader

def CreateSingleMtl(nodes, links, xmlPath):
    '''
    ### nodes : 节点组
    ### xmlPath: 查找节点位置
    '''
    bsdfShader = nodes.new("ShaderNodeBsdfPrincipled")

    NodeName =  ("Diffuse",     "Metallic", "Reflection",           "Roughness",    "Anisotropic",  "AnisotropicRotation",  "Sheen",        "SheenColor",   "SheenRoughness",   "Coat",         "CoatRoughness",   "IOR",  "Refraction",            "Emission",         "EmissionStrength",     "Opacity",  "Bump")
    inputName = ("Base Color",  "Metallic", "Specular IOR Level",   "Roughness",    "Anisotropic",  "Anisotropic Rotation", "Sheen Weight", "Sheen Tint",   "Sheen Roughness",  "Coat Weight",  "Coat Roughness",  "IOR",  "Transmission Weight",   "Emission Color",   "Emission Strength",    "Alpha",    "Normal")

    if bpy.app.version < (4, 0, 0):
        NodeName = ("Diffuse", "Metallic", "Reflection", "Roughness", "Anisotropic", "AnisotropicRotation", "Sheen", "Coat", "CoatRoughness", "IOR", "Refraction", "RefractRoughness", "Emission", "EmissionStrength", "Opacity", "Bump")
        inputName = ("Base Color", "Metallic", "Specular", "Roughness", "Anisotropic", "Anisotropic Rotation", "Sheen", "Clearcoat","Clearcoat Roughness", "IOR", "Transmission", "Transmission Roughness", "Emission", "Emission Strength", "Alpha", "Normal")

    useRoughness = xmlPath.attrib["useRoughness"]

    for path, input in zip(NodeName, inputName):
        node = xmlPath.find("./%s" % path)

        Default_Value = 1.0
        if "Color" in node.attrib :
            Default_Value = ConvertColor(node.attrib["Color"])
        
        if "Amount" in node.attrib :
            Default_Value = float(node.attrib["Amount"])
        
        texmap = node.attrib["Texmap"]
        texmap_type = texmap.split(":")[-1]
        texmap_node = xmlPath.find("./%s/" % path)

        parent_node = bsdfShader.inputs[input]  
        if path in ("Roughness", "CoatRoughness", "RefractRoughness", "SheenRoughness") :
            if useRoughness == "false" or path == "RefractRoughness":
                parent_node.default_value = 1 - Default_Value
                if texmap_node:
                    InvertNode = nodes.new("ShaderNodeInvert")
                    links.new(InvertNode.outputs["Color"], parent_node)
                    CreateNode(nodes, links, texmap_node, InvertNode.inputs["Color"], texmap_type)
        elif path in ("Bump"):
            if texmap_node:
                if texmap_node.tag in ("Normal"):
                    CreateNode(nodes, links, texmap_node, parent_node, texmap_type)
                else:
                    BumpNode = nodes.new("ShaderNodeBump")
                    BumpNode.inputs["Strength"].default_value = 0.1
                    links.new(BumpNode.outputs["Normal"], parent_node)
                    CreateNode(nodes, links, texmap_node, BumpNode.inputs["Height"], texmap_type)
        else:
            parent_node.default_value = Default_Value
            if texmap_node:
                CreateNode(nodes, links, texmap_node, parent_node, texmap_type)
    
    return bsdfShader

def import_max_scene_operator(**kwargs):    
    import_path = tempfile.gettempdir()

    MTLFile = os.path.join(import_path, "RH_M2B_TEMP", "RH_M2B.xml")
    FBXFile = os.path.join(import_path, "RH_M2B_TEMP", "RH_M2B.fbx")
    USDFile = os.path.join(import_path, "RH_M2B_TEMP", "RH_M2B.usd")

    ##### Importer Options #####
    import_model = kwargs["import_model"] if "import_model" in kwargs else False
    import_light = kwargs["import_light"] if "import_light" in kwargs else False
    import_proxy = kwargs["import_proxy"] if "import_proxy" in kwargs else False
    import_camera = kwargs["import_camera"] if "import_camera" in kwargs else False
    import_material = kwargs["import_material"] if "import_material" in kwargs else False
    import_settings = kwargs["import_settings"] if "import_settings" in kwargs else False
    import_environment = kwargs["import_environment"] if "import_environment" in kwargs else False
    import_animate = kwargs["import_animate"] if "import_animate" in kwargs else False

    use_collection = kwargs["use_collection"] if "use_collection" in kwargs else False
    toggle_cycles_render = kwargs["toggle_cycles_render"] if "toggle_cycles_render" in kwargs else False

    if os.path.exists(MTLFile) and (os.path.exists(FBXFile) or os.path.exists(USDFile)):
        xml_file = xml.parse(MTLFile)
        xml_root = xml_file.getroot()
        
        # Mesh Format
        MeshFormat = xml_root.find("./Setting/ExportFormat").text
        # Gamma
        gamma = float(xml_root.find("./Setting/Gamma").text)
        
        # 文件单位
        units = xml_root.find("./Setting/Units").text
        fac = 1 #缩放系数
        if units == "millimeters":
            fac = 0.001
        if units == "meters":
            fac = 1
        if units == "inches":
            fac = 0.0254
        if units == "feet":
            fac = 0.3048
        if units == "miles":
            fac == 1609.3439893783
        if units == "centimeters":
            fac = 0.01
        if units == "kilometers":
            fac = 1000
        
        # 错误材质列表
        error_mat_list = []
        
        print("============RED HALO Studio============")
        print("==============发霉的红地蛋==============")

        # 切换为Cycles渲染器
        if toggle_cycles_render :
            bpy.context.scene.render.engine = 'CYCLES'
        # 加载模型
        if import_model :
            print("========= 开始加载模型 Loading... =========")            

            try:
                if MeshFormat == "USD":
                    bpy.ops.wm.usd_import(filepath=USDFile)
                    # bpy.ops.wm.usd_import(filepath='', check_existing=False, filter_blender=False, filter_backup=False, filter_image=False, filter_movie=False, filter_python=False, filter_font=False, filter_sound=False, filter_text=False,
                    # filter_archive=False, filter_btx=False, filter_collada=False, filter_alembic=False, filter_usd=True, filter_obj=False, filter_volume=False, filter_folder=True, filter_blenlib=False, filemode=8, relative_path=True, 
                    # display_type='DEFAULT', sort_method='', filter_glob='*.usd', scale=1.0, set_frame_range=True, import_cameras=True, import_curves=True, import_lights=True, import_materials=True, import_meshes=True, import_volumes=True, 
                    # import_shapes=True, import_subdiv=False, import_instance_proxies=True, import_visible_only=True, create_collection=False, read_mesh_uvs=True, read_mesh_colors=True, prim_path_mask='', import_guide=False, import_proxy=True, 
                    # import_render=True, import_all_materials=False, import_usd_preview=True, set_material_blend=True, light_intensity_scale=1.0, mtl_name_collision_mode='MAKE_UNIQUE', import_textures_mode='IMPORT_PACK', import_textures_dir='//textures/', tex_name_collision_mode='USE_EXISTING')
                    for b in bpy.data.objects:
                        for c in b.constraints:
                            b.constraints.remove(c)
                    
                    for c in bpy.data.objects:
                        if c.type == "CURVE":
                            c.data.bevel_depth = 0

                if MeshFormat == "FBX":
                    use_anim = True if import_animate else False
                    bpy.ops.import_scene.fbx(filepath=FBXFile, use_custom_normals=False, use_custom_props=False, use_anim=use_anim, automatic_bone_orientation=True)
                    # bpy.ops.import_scene.fbx(filepath='', directory='', filter_glob='*.fbx', files=None, ui_tab='MAIN', use_manual_orientation=False, global_scale=1.0, bake_space_transform=False, use_custom_normals=True, use_image_search=True, 
                    # use_alpha_decals=False, decal_offset=0.0, use_anim=True, anim_offset=1.0, use_subsurf=False, use_custom_props=True, use_custom_props_enum_as_string=True, ignore_leaf_bones=False, force_connect_children=False, automatic_bone_orientation=False, 
                    # primary_bone_axis='Y', secondary_bone_axis='X', use_prepost_rot=True, axis_forward='-Z', axis_up='Y')

                objects_list = xml_root.find("./ObjectList")

                objects = [x.name for x in bpy.data.objects]

                for obj in objects_list.findall("./*") :
                    object_name = obj.attrib["Name"]
                    object_hash = obj.attrib["Hash"]

                    object_Render = True if obj.attrib["Render"] == "true" else False
                    object_camera = True if obj.attrib["Camera"] == "true" else False
                    object_reflect = True if obj.attrib["Reflect"] == "true" else False
                    object_shadow = True if obj.attrib["Shadow"] == "true" else False
                    object_atmospherics = True if obj.attrib["Atmospherics"] == "true" else False
                    
                    if object_hash not in objects:
                        continue 
                    
                    obj_scene = bpy.data.objects[object_hash]
                    if obj_scene.type == "MESH":
                        obj_scene.hide_render = not object_Render
                        # befor 3.00
                        if bpy.app.version < (3, 0, 0):
                            obj_vis = obj_scene.cycles_visibility
                            obj_vis.camera = object_camera
                            # obj_vis.diffuse = False
                            obj_vis.glossy = object_reflect
                            obj_vis.transmission = object_reflect
                            obj_vis.scatter = object_atmospherics
                            obj_vis.shadow = object_shadow

                        # 3.00 later
                        else:
                            obj_scene.visible_camera = object_camera
                            # obj_scene.visible_diffuse = False
                            obj_scene.visible_glossy = object_reflect
                            obj_scene.visible_transmission = object_reflect
                            obj_scene.visible_volume_scatter = object_atmospherics
                            obj_scene.visible_shadow = object_shadow

                    obj_scene.name = object_name
                    try:
                        obj_scene.data.name = object_name
                    except:
                        ...

            except:
                print(traceback.print_exc())
                pass
            
            print("========= 模型加载完成 Ended =========")
        
        # 把所有物体移动到新的集合中
        import_name, _ = os.path.splitext(xml_root.find("./Setting/File").text)
        Arch_Col = bpy.data.collections.new(import_name)
        bpy.context.collection.children.link(Arch_Col)

        objects_col = bpy.data.collections.new("Objects")
        Arch_Col.children.link(objects_col)

        for o in bpy.data.objects:
            objects_col.objects.link(o)
            bpy.context.collection.objects.unlink(o)

            if o.type == 'EMPTY' and o.children == ():
                bpy.data.objects.remove(o)

        # 设置代理模型
        if import_proxy :
            proxy_list = xml_root.find("./Proxy")

            proxy_col = bpy.data.collections.new("Proxy")
            Arch_Col.children.link(proxy_col)

            for p in proxy_list:
                source_obj_name = p.attrib["hash"]
                source_obj_file = p.attrib["filename"]
                _, tempfilename = os.path.split(source_obj_file)
                filename, extension = os.path.splitext(tempfilename)

                _src_obj = bpy.data.objects[source_obj_name]
                me = bpy.data.objects[source_obj_name].data
                
                empty_parent = bpy.data.objects.new((filename), None)
                proxy_col.objects.link(empty_parent)

                proxy_ins_item = p

                for idx, item in enumerate(proxy_ins_item):
                    item_name = item.attrib["name"]
                    item_matrix = ConvertMatrix(item.attrib["matrix"])

                    _tmp_obj = bpy.data.objects.new(item_name, me)
                    _tmp_obj.display_type = "BOUNDS"
                    _tmp_obj.matrix_world = item_matrix * fac

                    proxy_col.objects.link(_tmp_obj)
                    _tmp_obj.parent = empty_parent

                #Delete Origin objects
                bpy.data.objects.remove(_src_obj)

        # 设置自动平滑
        if bpy.app.version < (4, 1, 0):
            for i in bpy.data.meshes:
                i.use_auto_smooth = True

        # 设置场景
        if import_settings :
            renderWidth = int(xml_root.find("./Setting/Width").text)
            renderHeight = int(xml_root.find("./Setting/Height").text)
            bpy.context.scene.render.resolution_x = renderWidth
            bpy.context.scene.render.resolution_y = renderHeight

        # 设置相机
        if import_camera :
            camera_xml = xml_root.find("./CameraList")

            camera_col = bpy.data.collections.new("Camera")
            Arch_Col.children.link(camera_col)

            for c in camera_xml.findall("./*") :
                camera_name = c.attrib["name"]
                camera_matrix = ConvertMatrix(c.attrib["matrix"])
                camera_fov = c.attrib["fov"]
                camera_clip = c.attrib["clip"]
                camera_near = c.attrib["near"]
                camera_far = c.attrib["far"]

                camera_data = bpy.data.cameras.new(name='Cameras')
                camera_object = bpy.data.objects.new(camera_name, camera_data)
                camera_col.objects.link(camera_object)
                camera_object.data.angle = float(camera_fov)/180 * pi
                if camera_clip == "true":
                    camera_object.data.clip_start = float(camera_near) * fac
                    camera_object.data.clip_end = float(camera_far) * fac           

                camera_object.matrix_world = camera_matrix * fac
                camera_object.scale = (1, 1, 1)

        # 设置灯光
        if import_light :
            print("========= 加载灯光设置 Loading... =========")

            light_col = bpy.data.collections.new("Lights")
            Arch_Col.children.link(light_col)
            light_xml = xml_root.find("./LightList")
            for x in light_xml.findall("./*"):
                L_Type = x.attrib["type"]
                IES = x.attrib["ies"]
                light_name = "tmp_light_name"
                color = ConvertColor(x.attrib["color"])

                if L_Type in ("AREA", "DISK"):
                    light_data = bpy.data.lights.new(name="Light", type = "AREA")
                else:
                    light_data = bpy.data.lights.new(name="Light", type = L_Type)

                light_obj = bpy.data.objects.new(light_name, light_data)
                light_obj.data.energy = float(x.attrib["multiplier"])
                light_obj.data.color = color[:-1]

                if L_Type == "POINT":
                    light_obj.data.shadow_soft_size = 0.005 if float(x.attrib["length"]) * fac < 0.005 else float(x.attrib["length"]) * fac                        
                elif L_Type == "AREA":
                    light_obj.data.shape = 'RECTANGLE'
                    light_obj.data.size = 0.005 if float(x.attrib["length"]) * fac < 0.005 else float(x.attrib["length"]) * fac
                    light_obj.data.size_y = 0.005 if float(x.attrib["width"]) * fac < 0.005 else float(x.attrib["width"]) * fac
                    light_obj.data.cycles.is_portal = True if x.attrib["portal"] == "true" else False
                elif L_Type == "DISK":
                    light_obj.data.shape = 'DISK'
                    light_obj.data.size = 0.005 if float(x.attrib["length"]) * fac < 0.005 else float(x.attrib["length"]) * fac
                    light_obj.data.size_y = 0.005 if float(x.attrib["width"]) * fac < 0.005 else float(x.attrib["width"]) * fac
                    light_obj.data.cycles.is_portal = True if x.attrib["portal"] == "true" else False
                elif L_Type == "SUN":
                    light_obj.data.angle = 0.005
                elif L_Type == "SPOT" :
                    pass
                
                ### Add IES File
                if IES != "undefined":
                    light_obj.data.use_nodes = True

                    nodes = light_obj.data.node_tree.nodes
                    links = light_obj.data.node_tree.links

                    IES_Node = nodes.new("ShaderNodeTexIES")
                    IES_Node.mode = 'EXTERNAL'
                    IES_Node.filepath = IES
                    IES_Node.location = (-220, 300)
                    Emission_Node = nodes["Emission"]
                    links.new(IES_Node.outputs[0], Emission_Node.inputs["Strength"])

                for light in x.findall("./*") :
                    matrix = ConvertMatrix(light.attrib["matrix"])

                    light_instance = light_obj.copy()
                    light_instance.name = light.attrib["name"]
                    light_instance.matrix_world = matrix * fac
                    light_instance.scale *= 1 / fac

                    light_instance.visible_camera = True if light.attrib["invisible"] == "false" else False
                    light_instance.visible_diffuse = True if light.attrib["affectdiffuse"] == "true" else False
                    light_instance.visible_glossy = True if light.attrib["affectspecular"] == "true" else False
                    light_instance.visible_transmission = True if light.attrib["affectreflections"] == "true" else False
                    
                    light_col.objects.link(light_instance)
                
                #Delete Origin objects
                bpy.data.objects.remove(light_obj)

            print("========= 灯光设置完成 Ended =========")

        # 设置环境
        '''
        if import_environment :
            print("========= 加载环境设置 Loading... =========")

            env_color = ConvertColor(xml_root.find("./Environment/Color").text)
            env_intensity = float(xml_root.find("./Environment/Multiplier").text)
            env_texmap = xml_root.find("./Environment/Texmap")

            scene = bpy.context.scene
            world = scene.world
            world.use_nodes = True
            nodes = world.node_tree.nodes
            links = world.node_tree.links

            # clear default nodes
            nodes.clear()

            # create input node
            world_output_node = nodes.new("ShaderNodeOutputWorld")
            world_output_node.location = (0, 0)

            backgroud_node = nodes.new("ShaderNodeBackground")
            backgroud_node.location = (-200, 0)
            backgroud_node.inputs[0].default_value = env_color
            backgroud_node.inputs[1].default_value = env_intensity

            links.new(backgroud_node.outputs[0], world_output_node.inputs[0])

            # create texmap node
            if env_texmap :                
                if env_texmap.attrib["Texmap"] != "undefined":
                    texmap_type = env_texmap.attrib["Texmap"].split(":")[-1]
                    CreateNode(nodes, links, env_texmap[0], backgroud_node.inputs[0], texmap_type)
                
            print("========= 环境设置完成 Ended =========")
        
        '''

        # 设置材质
        if import_material :
            print("========= 加载材质 Loading... =========")
            SceneMaterials = [ m for m in bpy.data.materials[:] if not m.is_grease_pencil]
            
            Create_ColorCorrection()

            for mat in SceneMaterials:
                material_name = mat.name
                nodes = mat.node_tree.nodes
                links = mat.node_tree.links
                
                mat.blend_method = "CLIP"
                mat.shadow_method = "HASHED"
                mat.metallic = 0

                m = xml_root.find("./MaterialList/*/[@Hash='%s']" % material_name)
                try:
                    nodes.clear()
                    outputShader = nodes.new("ShaderNodeOutputMaterial")
                    if hasattr(m, "tag"):
                        origin_name = m.attrib["Name"]
                        mat.name = origin_name
                        ### Double Materials
                        if m.tag == "VRayDoubleSideMtl" :
                            DoubleMixShader = nodes.new("ShaderNodeMixShader")
                            FrontBSDFShader = None
                            BackBSDFShader = None
                            TranslucentShader = None

                            links.new(DoubleMixShader.outputs["Shader"], outputShader.inputs["Surface"])
                            
                            frontMtl = m.attrib["frontmtl"].split(":")[0]
                            backMtl = m.attrib["backmtl"].split(":")[0]
                            mix_value = float(m.attrib["mix"])
                            
                            mask_map = m.attrib["mask"]                            
                            force1sidedsubmtls = True if m.attrib["force1sidedsubmtls"] == "true" else False
                            
                            # backmtl_on = m.attrib["backmtlon"]

                            DoubleMixShader.inputs["Fac"].default_value = mix_value

                            if frontMtl != "undefined":
                                FMat = xml_root.find("./MaterialList/SingleMtl/[@Hash='%s']" % frontMtl)
                                FrontBSDFShader = CreateSingleMtl(nodes, links, FMat)
                                
                            if force1sidedsubmtls:
                                TranslucentShader = nodes.new("ShaderNodeBsdfTranslucent")
                                links.new(TranslucentShader.outputs["BSDF"], DoubleMixShader.inputs[2])
                                links.new(FrontBSDFShader.outputs["BSDF"], DoubleMixShader.inputs[1])
                                
                            else:
                                transparentShader = nodes.new("ShaderNodeBsdfTransparent")
                                links.new(transparentShader.outputs["BSDF"], DoubleMixShader.inputs[2])

                                DoubleMixBackShader = nodes.new("ShaderNodeMixShader")
                                links.new(DoubleMixBackShader.outputs["Shader"], DoubleMixShader.inputs[1])

                                backfacingNode = nodes.new("ShaderNodeNewGeometry")
                                for o in backfacingNode.outputs:
                                    if o.name != "Backfacing":
                                        o.hide = True
                                links.new(backfacingNode.outputs["Backfacing"], DoubleMixBackShader.inputs["Fac"])
                                links.new(FrontBSDFShader.outputs["BSDF"], DoubleMixBackShader.inputs[1])

                                if backMtl != "undefined":
                                    if backMtl == frontMtl:
                                        links.new(FrontBSDFShader.outputs["BSDF"], DoubleMixBackShader.inputs[2])
                                    else:
                                        BMat = xml_root.find("./MaterialList/SingleMtl/[@Hash='%s']" % backMtl)
                                        BackBSDFShader = CreateSingleMtl(nodes, links, BMat)
                                        links.new(BackBSDFShader.outputs["BSDF"], DoubleMixBackShader.inputs[2])
                            
                            if mask_map != "undefined":
                                texmap_type = m.attrib["mask"].split(":")[-1]
                                CreateNode(nodes, links, m[0], DoubleMixShader.inputs["Fac"], texmap_type)

                                # Find base color of front mtl, and link to transnode                                
                                for lk in links:
                                    bc_from = None
                                    alpha_from = None
                                    if lk.to_socket.name == "Base Color":
                                        bc_from = lk.from_node
                                        links.new(bc_from.outputs[0], TranslucentShader.inputs["Color"])

                                    # Alpha ???
                                    if lk.to_socket.name == "Alpha":
                                        alpha_from = lk.from_node
                                        # 旧版MixRGB
                                        # if bpy.app.version < (3,5,0):
                                        # pass
                                        MixShader = nodes.new("ShaderNodeMixRGB")
                                        # MixShader = nodes.new("ShaderNodeMix")
                                        MixShader.blend_type = 'MULTIPLY'
                                        MixShader.inputs[0].default_value = 1

                                        links.new(alpha_from.outputs[0], MixShader.inputs["Color1"])
                                        if bc_from != None:
                                            links.new(bc_from.outputs[0], MixShader.inputs["Color2"])
                                        links.new(MixShader.outputs[0], TranslucentShader.inputs["Color"])
                        
                        elif m.tag == "DoubleMtl": 
                            DoubleMixShader = nodes.new("ShaderNodeMixShader")
                            FrontBSDFShader = None
                            BackBSDFShader = None
                            TranslucentShader = None

                            links.new(DoubleMixShader.outputs["Shader"], outputShader.inputs["Surface"])
                            
                            frontMtl = m.attrib["mtl1"].split(":")[0]
                            backMtl = m.attrib["mtl2"].split(":")[0]
                            mix_value = float(m.attrib["mix1"])
                            
                            mask_map = m.attrib["mask1"]                            

                            DoubleMixShader.inputs["Fac"].default_value = mix_value

                            if frontMtl != "undefined":
                                FMat = xml_root.find("./MaterialList/SingleMtl/[@Hash='%s']" % frontMtl)
                                FrontBSDFShader = CreateSingleMtl(nodes, links, FMat)
                                links.new(FrontBSDFShader.outputs["BSDF"], DoubleMixShader.inputs[1])
                                
                            if backMtl != "undefined":
                                if backMtl == frontMtl:
                                    links.new(FrontBSDFShader.outputs["BSDF"], DoubleMixShader.inputs[2])
                                else:
                                    BMat = xml_root.find("./MaterialList/SingleMtl/[@Hash='%s']" % backMtl)
                                    BackBSDFShader = CreateSingleMtl(nodes, links, BMat)
                                    links.new(BackBSDFShader.outputs["BSDF"], DoubleMixShader.inputs[2])
                            
                            if mask_map != "undefined":
                                texmap_type = m.attrib["mask1"].split(":")[-1]
                                CreateNode(nodes, links, m[0], DoubleMixShader.inputs["Fac"], texmap_type)

                        ### Single Material
                        elif m.tag == "SingleMtl" :
                            node = CreateSingleMtl(nodes, links, m)
                            # TranslucecyNode
                            # 1.判断是不是有 Tanslucecy节点
                            # 2.判断是不是有 Alpha 节点，如果有的话，就把Alpha和Tanslucecy节点混合再连入Translucecy节点中
                            translucency_path = m.find("./Translucent")
                            alpha_path = m.find("./Opacity")
                            
                            if translucency_path.attrib["Texmap"] != "undefined" :
                                # AddShader = nodes.new("ShaderNodeAddShader")
                                AddShader = nodes.new("ShaderNodeMixShader")
                                TranslucentShader = nodes.new("ShaderNodeBsdfTranslucent")
                                AddShader.inputs["Fac"].default_value = 0.4

                                links.new(AddShader.outputs["Shader"], outputShader.inputs["Surface"])
                                links.new(TranslucentShader.outputs["BSDF"], AddShader.inputs[2])
                                links.new(node.outputs["BSDF"], AddShader.inputs[1])

                                texmap_type = translucency_path.attrib["Texmap"].split(":")[-1]

                                if alpha_path.attrib["Texmap"] != "undefined" :
                                    # 旧版MixRGB
                                    MultiplyNode = nodes.new("ShaderNodeMixRGB")

                                    # 新版MixRGB
                                    # MultiplyNode = nodes.new("ShaderNodeMix")
                                    MultiplyNode.blend_type = 'MULTIPLY'

                                    # 新版Mix参数
                                    # MultiplyNode.data_type = 'RGBA'
                                    MultiplyNode.inputs[0].default_value = 1
                                    
                                    # 旧版Mix参数
                                    links.new(MultiplyNode.outputs["Color"], AddShader.inputs["Fac"])
                                    ## Add Mode
                                    links.new(MultiplyNode.outputs["Color"], TranslucentShader.inputs["Color"])
                                    # 新版Mix参数
                                    # links.new(MultiplyNode.outputs[2], TranslucentShader.inputs["Color"])

                                    for l in node.inputs["Alpha"].links:
                                        links.new(l.from_socket, MultiplyNode.inputs["Color2"])
                                        # links.new(l.from_socket, MultiplyNode.inputs[6])

                                    # 旧版Mix连接方式
                                    CreateNode(nodes, links, translucency_path[0], MultiplyNode.inputs["Color1"], texmap_type)
                                    # 新版Mix连接方式
                                    # CreateNode(nodes, links, translucency_path[0], MultiplyNode.inputs[7], texmap_type)
                                else:
                                    CreateNode(nodes, links, translucency_path[0], TranslucentShader.inputs["Color"], texmap_type)
                                    CreateNode(nodes, links, translucency_path[0], AddShader.inputs["Fac"], texmap_type)
                            else:
                                links.new(node.outputs["BSDF"], outputShader.inputs["Surface"])

                            # Displacement
                            Displacement_path = m.find("./Displacement")
                            if Displacement_path.attrib["Texmap"] != "undefined" :
                                DisplacementShader = nodes.new("ShaderNodeDisplacement")
                                DisplacementShader.mute = True
                                texmap_type = Displacement_path.attrib["Texmap"].split(":")[-1]

                                CreateNode(nodes, links, Displacement_path[0], DisplacementShader.inputs["Normal"], texmap_type)
                                links.new(DisplacementShader.outputs["Displacement"], outputShader.inputs["Displacement"])
                            
                            # 禁用Displacement连线
                            for l in links:
                                if l.to_socket.name == "Displacement":
                                    l.is_muted = True

                        ### Light Material
                        elif m.tag == "LightMtl" :
                            emissionShader = nodes.new("ShaderNodeEmission")
                            
                            doubleSideShader = nodes.new("ShaderNodeMixShader")
                            mixTransShader = nodes.new("ShaderNodeMixShader")
                            TransparentShader = nodes.new("ShaderNodeBsdfTransparent")
                            GeometryShader = nodes.new("ShaderNodeNewGeometry")
                            MathShader = nodes.new("ShaderNodeMath")

                            links.new(mixTransShader.outputs["Shader"], outputShader.inputs["Surface"])
                            links.new(GeometryShader.outputs["Backfacing"], MathShader.inputs[0])
                            links.new(MathShader.outputs["Value"], doubleSideShader.inputs["Fac"])
                            links.new(emissionShader.outputs["Emission"], doubleSideShader.inputs[1])
                            links.new(doubleSideShader.outputs["Shader"], mixTransShader.inputs[1])
                            links.new(TransparentShader.outputs["BSDF"], doubleSideShader.inputs[2])
                            links.new(TransparentShader.outputs["BSDF"], mixTransShader.inputs[2])

                            emissionShader.location = (-750, 0)
                            mixTransShader.location = (-250, 0)
                            doubleSideShader.location = (-500, 0)
                            MathShader.location = (-750, 200)
                            GeometryShader.location = (-1000, 300)
                            TransparentShader.location = (-750, -200)

                            # 设置参数
                            emissionShader.inputs["Color"].default_value = ConvertColor(m.attrib["color"])
                            emissionShader.inputs["Strength"].default_value = float(m.attrib["multiplier"])

                            mixTransShader.inputs["Fac"].default_value = 0
                            MathShader.operation = 'MULTIPLY'
                            if m.attrib["twoSided"] == "false":
                                MathShader.inputs[1].default_value = 1
                            else :
                                MathShader.inputs[1].default_value = 0

                            if m.attrib["texmap"] != "undefined":                            
                                texmap_type = m.attrib["texmap"].split(":")[-1]
                                CreateNode(nodes, links, m[0][0], emissionShader.inputs["Color"], texmap_type)
                            
                            if m.attrib["opacity_texmap"] != "undefined" :
                                texmap_type = m.attrib["texmap"].split(":")[-1]
                                CreateNode(nodes, links, m[1][0], mixTransShader.inputs["Fac"], texmap_type)

                except Exception as ex:
                    print(traceback.print_exc())
                    error_mat_list.append(material_name)

            ## Set all images colorspace is Non-Color
            for i in bpy.data.images :
                i.colorspace_settings.name = "Non-Color" # "sRGB"  
            
            ## Set images ColorSapce
            img_nodes = []
            for mat in SceneMaterials:
                nodes = mat.node_tree.nodes
                links = mat.node_tree.links

                for i in nodes:
                    if i.type in ("BSDF_PRINCIPLED", "EMISSION"):
                        for n in ("Base Color", "Subsurface Color", "Emission", "Color"):
                            try:
                                nn = i.inputs[n]
                                if nn.is_linked:
                                    for l in nn.links:
                                        img_nodes.extend(CollectImagesNode(l.from_node))
                            except:
                                pass
            for i in img_nodes:
                if i.type == "TEX_IMAGE":
                    i.image.colorspace_settings.name = "sRGB"                    
        
        print("========= 材质加载完成 Ended =========")
        print("=========== RED HALO Studio =========")
        # 错误材质列表
        return error_mat_list
    else:
        return [0]

class OT_MaxFileBrower(Operator, ImportHelper):
    bl_idname = "redhalo.maxfilebrower"
    bl_label = "RedHaloM2B v." + __version__
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_options = {'UNDO'}
    
    filter_glob: StringProperty(
        default='*.max',
        options={'HIDDEN'}
    )

    import_model: BoolProperty(
        name = "Mesh",
        description = "Import Mesh",
        default = True
    )

    import_light: BoolProperty(
        name = "Light",
        description = "Import Light",
        default = True
    )

    import_proxy: BoolProperty(
        name = "Proxy",
        description = "Import Proxy",
        default = True
    )

    import_camera: BoolProperty(
        name = "Camera",
        description = "Import Camera",
        default = True
    )

    import_material: BoolProperty(
        name = "Material",
        description = "Import Material",
        default = True
    )

    import_settings: BoolProperty(
        name = "Settings",
        description = "Import Settings",
        default = True
    )

    use_collection: BoolProperty(
        name = "Use Collection",
        description = "Use Collection",
        default = True
    )
    
    import_animate: BoolProperty(
        name = "Animate",
        description = "Import Animate",
        default = True
    )
    
    onlyImport: BoolProperty(
        name = "Only Import",
        description = "Only Import",
        default = False
    )

    import_format: EnumProperty(
        name = "Format",
        description = "Export Scene Type",
        #[(identifier, name, description, icon, number), ...]
        items=[("USD", "Fast", "导入速度最快，但是软件版本要求高", 'FORCE_FORCE', 1),
               ("FBX", "Slow", "兼容性最高，但是导入速度很慢", 'SORTTIME', 2),
        ],
        default="USD",
    )

    toggle_cycles_render : BoolProperty(
        name = "Cycles Render",
        description = "Change to Cycles Render",
        default = True
    )
    
    def execute(self, context):
        pref = context.preferences
        addon_pref = pref.addons[__package__].preferences
        maxroot = addon_pref.MaxRootFolder

        script_file = os.path.realpath(__file__)
        directory = os.path.dirname(script_file)

        ini_file = os.path.join(directory, "RH_M2B.ini")
        version = bpy.app.version_string
        version = version.split(".")
        version_m = version[0] 
        version_b = version[1]

        # Write profile
        with open(ini_file, "w") as f:
            f.write("[RH_M2B]\n")
            f.write("main={}\n".format(version_m))
            f.write("branch={}\n".format(version_b))

            if bpy.app.version < (3, 5, 0):
                f.write("format=FBX")
            else:
                f.write("format={}".format(self.import_format))

            f.close
        
        # Set New DateName option to False
        pref.view.use_translate_new_dataname = False

        if not self.onlyImport:            
            maxcmd = maxroot + "3dsmaxcmd.exe"
            maxfile = self.filepath
            maxscript = os.path.join(directory, "export.mse")
            extra = '-script:"' + maxscript + '"' #"  -secure:0
            doscmd = ('\""{}" "{}" {}\"').format(maxcmd, maxfile, extra)

            if os.path.exists(maxcmd):
                ## Max Applicate Version
                MaxAppVersion = ""
                with open(maxcmd, "rb") as f:
                    buff = f.read()
                    inx = buff.find(b"\x01\x00\x46\x00\x69\x00\x6C\x00\x65\x00\x56\x00\x65\x00\x72\x00\x73\x00\x69\x00\x6F\x00\x6E\x00\x00\x00\x00\x00")
                    f.seek(inx+28)
                    temp = f.read(24).decode("utf-8")
                    for i in range(0, len(temp), 2):
                        MaxAppVersion += temp[i]
                        
                MaxAppVersion = int(MaxAppVersion.split(".")[0])
                
                ## Max File Version
                MaxFileVerion = FindMax(maxfile)

                if MaxFileVerion > MaxAppVersion:
                    self.report({'ERROR'}, "文件版本太高，请换一个更高版本的Max程序")
                    return {'CANCELLED'}
                else:
                    os.system(doscmd)
            else:
                self.report({'ERROR'}, "Max程序目录不正确，请重新指定")
                return {'CANCELLED'}
        
        
        # Import Scene
        err = import_max_scene_operator(
            import_model = self.import_model, 
            import_light = self.import_light, 
            import_proxy = self.import_proxy, 
            import_camera = self.import_camera, 
            import_material = self.import_material, 
            import_settings = self.import_settings, 
            use_collection = self.use_collection,
            toggle_cycles_render = self.toggle_cycles_render
        )

        if len(err) > 0:
            if err[0] == 0:
                self.report({'ERROR'}, "文件无法转换，错误日志位于 %temp%\\RH_M2B_TEMP")
            else:
                self.report({'ERROR'}, "材质不能完整转换,详细列表看控制台")
    
            print("\t {} 个材质不能转换，名称如下".format(len(err)))
            for i,m in enumerate(err):
                print("\t{}/{}:{}".format(i+1, len(err),m))
              
        return {'FINISHED'}
    
    def draw(self, context):
        layout = self.layout
        
        box = layout.box()
        box.label(text = "依赖条件：", icon='ERROR')
        box.label(text = "1.3dsMax主程序（越高越好）")
        box.label(text = "2.对应材质相关的渲染器：")
        box.label(text = "  a.VRay英文版(5.0+)")
        box.label(text = "  b.Corona英文版")

        layout.label(text = "Import format:")
        row = layout.row()
        row.scale_y = 1.25
        row.prop(self, "import_format", expand = True)

        layout.label(text = "Import Options:")
        box = layout.box()
        row = box.row(align=True)
        row.scale_y = 1.25
        row.prop(self, "import_model", icon="MESH_DATA")
        row.prop(self, "import_light", icon="LIGHT_DATA")
        row = box.row(align=True)
        row.scale_y = 1.25
        row.prop(self, "import_proxy", icon="LINK_BLEND")
        row.prop(self, "import_camera", icon="CAMERA_DATA")
        row = box.row(align=True)
        row.scale_y = 1.25
        row.prop(self, "import_material", icon="MATERIAL")
        row.prop(self, "import_settings", icon="PREFERENCES")
        
        layout.label(text = "Others:")
        box = layout.box()
        row = box.row(align=True)
        row.scale_y = 1.25
        row.prop(self, "toggle_cycles_render", icon="RESTRICT_RENDER_OFF")

        layout.label(text = "Import Options:")        
        box = layout.box()
        box.scale_y = 2
        box.prop(self, "onlyImport", icon="IMPORT")

class REDHALO_M2B_PREFS(AddonPreferences):
    bl_idname = __package__

    MaxRootFolder: StringProperty(
		name = "3dsMax Root",
		description = "select 3ds max install folder",
		subtype = 'DIR_PATH',
	)

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        box.label(text = "一定要正确选择max的安装路径，不然不能正常使用")
        box.prop(self, "MaxRootFolder")

        box = layout.box()
        box.label(text = "此插件依赖的max程序，只是做的后台自动导出和导入，因此max主程序还是要安装滴。")
        box.label(text = "关于渲染器：")
        box.label(text = "如果场景材质是VRay材质，那么必须安装VRay英文版")
        box.label(text = "如果场景材质是Corona材质，那么必须安装Corona英文版")
        box.label(text = "由于高版的VRay和Corona部分材质可以互相使用，因此建议2个渲染都安装")
        box.label(text = "其它：")
        box.label(text = "1.导出的文件位置系统的临时文件夹【%temp%】中")
        box.label(text = "2.对于安装3dsMax安装完成后C盘占用太大的，可以去控制面板把名字带有Autodesk Material相关的安装程序全部卸载.")
        
        box = layout.box()
        box.label(text = "支持的材质：")
        box.label(text = "VRay标准材质，VRay灯光材质，VRay双面材质，Corona标准材质，Corona物理材质，Corona灯光材质，标准材质")
        box.label(text = "支持的纹理")
        box.label(text = "Bitmap, Mix, AO, Brick(部分), Checker, Falloff(部分), ColorCorrection, RGB Multiply, Gradient, Normal, VRayBitmap, VrayColor, CoronaMix, CornaFrontBack等等")
        box.label(text = "支持的材质和纹理")

def menu_func_import(self, context):
    self.layout.operator(OT_MaxFileBrower.bl_idname, text="3ds Max File Importer(.max)", icon="EVENT_M")

classes = (
    OT_MaxFileBrower,
    REDHALO_M2B_PREFS
)

def register():
    for cls in classes:
        bpy.utils.register_class(cls)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import)

def unregister():
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import)    
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)